<?php

/**
 * @file
 * Authenticate keyed requests to REST Easy APIs.
 */

namespace Drupal\rest_easy_api_keys\Authentication;

use Drupal\Core\Authentication\AuthenticationProviderInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\rest_easy\Plugin\APIManager;
use Drupal\rest_easy\Plugin\EndpointBase;
use Drupal\user\Entity\User;
use Symfony\Component\HttpFoundation\Request;

/**
 * Class RESTEasyAPIKeyAuth
 *
 * @package Drupal\rest_easy_api_keys\Authentication
 */
class RESTEasyAPIKeyAuth implements AuthenticationProviderInterface {

  /**
   * @var array
   *   A list of paths for REST Easy endpoints using this authentication method
   */
  protected $paths = [];

  /**
   * @var \Drupal\Core\Entity\Query\QueryInterface
   *   User query service supplied by the Drupal container
   */
  protected $user_query;

  /**
   * APIKeyAuth constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function __construct(APIManager $api_manager, EntityTypeManagerInterface $entity_type_manager) {
    foreach ($api_manager->getDefinitions() as $api_id => $api_definition) {
      try {
        $api = $api_manager->createInstance($api_id);
        foreach ($api->routes() as $route) {
          $auth_plugins = $route->getOption('_auth');
          if ($auth_plugins && in_array('rest_easy_api_key', $auth_plugins)) {
            $path = $route->getPath();
            if (strpos($path, '{') !== FALSE) {
              $path = substr($path, 0, strpos($path, '{'));
            }
            $this->paths[] = $path;
          }
        }
      }
      catch (\Exception $exception) {
        continue;
      }
    }
    $this->user_query = $entity_type_manager->getStorage('user')->getQuery();
  }

  /**
   * {@inheritdoc}
   */
  public function applies(Request $request) {

    // Check if the request is for an endpoint that uses this authentication method.
    $in_paths = FALSE;
    $current_path = $request->getRequestUri();
    foreach ($this->paths as $path) {
      if (str_starts_with($current_path, $path)) {
        $in_paths = TRUE;
        break;
      }
    }
    if (!$in_paths) {
      return FALSE;
    }

    // Check if the request has an API key.
    return $request->query->has('apiKey') || $request->headers->has('x-api-key');
  }

  /**
   * {@inheritdoc}
   */
  public function authenticate(Request $request) {
    $api_key = $request->query->get('apiKey') ?: $request->headers->get('x-api-key');
    if (!$api_key) {
      EndpointBase::status(403);
    }
    $user = $this->user_query
      ->accessCheck(FALSE)
      ->condition('rest_easy_api_keys', $api_key)
      ->execute();
    if (count($user) > 0) {
      $user = User::load(reset($user));
    }
    if (!$user) {
      EndpointBase::status(403);
    }
    return $user;
  }
}
