<?php

/**
 * @file
 * Define the methods every API plugin must implement.
 */

namespace Drupal\rest_easy\Plugin;

use Drupal\Component\Plugin\PluginInspectionInterface;

/**
 * Interface APIPluginInterface
 *
 * @package Drupal\rest_easy\Plugin
 */
interface APIInterface extends PluginInspectionInterface {

  /**
   * Retrieve the API name for OpenAPI documentation.
   *
   * @return string
   *   The API name
   */
  public function getApiName(): string;

  /**
   * Retrieve the API base path for OpenAPI documentation.
   *
   * @return string
   *   The API base path
   */
  public function getBasePath(): string;

  /**
   * Retrieve the API description for OpenAPI documentation.
   *
   * @return string
   */
  public function getDescription(): string;

  /**
   * Generate API information for OpenAPI documentation.
   *
   * @return array
   *   The API information
   */
  public function getInfo(): array;

  /**
   * Retrieve the JSON schema for OpenAPI documentation.
   *
   * @param $described_format
   * @param $entity_type_id
   * @param $bundle_name
   *
   * @return array
   * 
   * @see \Drupal\rest_easy\Plugin\openapi\OpenApiGenerator\RESTEasyApiGenerator::getJsonSchema()
   */
  public function getJsonSchema($described_format, $entity_type_id, $bundle_name = NULL);

  /**
   * Assemble a list of shared parameters for OpenAPI documentation.
   *
   * @return array
   *   An associative array of shared parameters
   * 
   * @see \Drupal\rest_easy\Plugin\openapi\OpenApiGenerator\RESTEasyApiGenerator::getSpecification()
   */
  public function getParameters(): array;

  /**
   * Assemble paths for OpenAPI documentation.
   *
   * @return array
   */
  public function getPaths(): array;

  /**
   * Assemble schema definitions for OpenAPI documentation.
   *
   * @return array[]
   * 
   * @see \Drupal\rest_easy\Plugin\openapi\OpenApiGenerator\RESTEasyApiGenerator::getSchemaDefinitions()
   */
  public function getSchemaDefinitions(): array;

  /**
   * Assemble security definitions for OpenAPI documentation.
   * 
   * @return array
   * 
   * @see \Drupal\rest_easy\Plugin\openapi\OpenApiGenerator\RESTEasyApiGenerator::getSecurityDefinitions()
   */
  public function getSecurityDefinitions(): array;

  /**
   * Generate API tags for OpenAPI documentation.
   *
   * @return array
   *   An array of tag names and descriptions
   */
  function getTags(): array;

  /**
   * Generate routes for the API.
   *
   * @return array
   *
   * @see \Drupal\rest_easy\Routing\APIRouteBuilder::routes()
   */
  function routes(): array;
}
